/*
 *  gensiomux - A program for connecting gensios.
 *  Copyright (C) 2020  Mats Webjörn <konsult@webjorn.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <gensio/gensio.h>
#include <gensio/gensio_class.h>
#include <gensio/gensio_selector.h>
#include <gensio/gensio_builtins.h>
#include <gensio/gensio_list.h>

#include "ioinfo.h"
#include "ser_ioinfo.h"
#include "utils.h"
#include "localports.h"

unsigned int debug;

#ifdef HAVE_OPENSSL
/*
 * Set a dummy random upstream file, for reproducable openssl usage for
 * fuzz testing.
 */
#include <openssl/rand.h>

static int dummyrnd_file = -1;

int dummyrnd_seed(const void *buf, int num)
{
	return 1;
}

int dummyrnd_bytes(unsigned char *buf, int num)
{
	int rc;
	int count = 0;

	while (num > 0)
	{
		rc = read(dummyrnd_file, buf, num);
		if (rc < 0)
		{
			fprintf(stderr, "Error reading from dummyrnd file: %s\n",
					strerror(errno));
			return 0;
		}
		if (rc == 0)
		{
			off_t pos = lseek(dummyrnd_file, 0, SEEK_SET);

			if (pos == -1)
			{
				fprintf(stderr, "Error seeking dummyrnd file: %s\n",
						strerror(errno));
				return 0;
			}
			rc = read(dummyrnd_file, buf, num);
			if (rc <= 0)
			{
				fprintf(stderr, "Error reading from dummyrnd file: %s\n",
						strerror(errno));
				return 0;
			}
		}
		count += rc;
		buf += rc;
		num -= rc;
	}

	return count;
}

void dummyrnd_cleanup(void)
{
}

int dummyrnd_add(const void *buf, int num, double randomness)
{
	return 1;
}

int dummyrnd_pseudorand(unsigned char *buf, int num)
{
	return dummyrnd_bytes(buf, num);
}

int dummyrnd_status(void)
{
	return 1;
}

struct rand_meth_st dummyrnd = {
	.seed = dummyrnd_seed,
	.bytes = dummyrnd_bytes,
	.cleanup = dummyrnd_cleanup,
	.add = dummyrnd_add,
	.pseudorand = dummyrnd_pseudorand,
	.status = dummyrnd_status,
};
#endif

struct gdata
{
	struct gensio_os_funcs *o;
	struct gensio_waiter *waiter;

	struct gensio *user_io;
	struct gensio *io;
	const char *ios;
	bool can_close;
	int err;

	struct gensio_list cons;
};

static void
gshutdown(struct ioinfo *ioinfo, bool user_req)
{
	struct gdata *ginfo = ioinfo_userdata(ioinfo);

	if (!user_req)
		ginfo->err = 1;
	ginfo->o->wake(ginfo->waiter);
}

static void
gerr(struct ioinfo *ioinfo, char *fmt, va_list ap)
{
	struct gdata *ginfo = ioinfo_userdata(ioinfo);

	fprintf(stderr, "Error on %s: ", ginfo->ios);
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\r\n");
}

static void
gout(struct ioinfo *ioinfo, char *fmt, va_list ap)
{
	struct gdata *ginfo = ioinfo_userdata(ioinfo);
	char str[200];

	vsnprintf(str, sizeof(str), fmt, ap);
	gensio_write(ginfo->user_io, NULL, str, strlen(str), NULL);
}

static struct ioinfo_user_handlers guh = {
	.shutdown = gshutdown,
	.err = gerr,
	.out = gout};

static void
io_open(struct gensio* io, int err, void* open_data)
{
	struct ioinfo* ioinfo = gensio_get_user_data(io);
	struct gdata*  ginfo  = ioinfo_userdata(ioinfo);

	if (err)
	{
		ginfo->can_close = false;
		fprintf(stderr, "open error on %s: %s\n", ginfo->ios,
				gensio_err_to_str(err));
		gshutdown(ioinfo, false);
	}
	else
	{
		ioinfo_set_ready(ioinfo, io);
	}
}

static void
io_close(struct gensio *io, void *close_data)
{
	struct ioinfo *ioinfo = gensio_get_user_data(io);
	struct gdata *ginfo = ioinfo_userdata(ioinfo);
	struct gensio_waiter *closewaiter = close_data;

	ginfo->o->wake(closewaiter);
}

static int
io_acc_event(struct gensio_accepter *accepter, void *user_data,
			 int event, void *data)
{
	struct ioinfo *ioinfo = user_data;
	struct gdata *ginfo = ioinfo_userdata(ioinfo);
	struct ioinfo *oioinfo = ioinfo_otherioinfo(ioinfo);
	struct gdata *oginfo = ioinfo_userdata(oioinfo);
	int rv;

	if (event == GENSIO_ACC_EVENT_LOG)
	{
		struct gensio_loginfo *li = data;

		vfprintf(stderr, li->str, li->args);
		fprintf(stderr, "\n");

		ginfo->err = 1;
		ginfo->o->wake(ginfo->waiter);
		return 0;
	}

	if (event != GENSIO_ACC_EVENT_NEW_CONNECTION)
		return GE_NOTSUP;

	if (ginfo->io)
	{
		gensio_free(data);
		return 0;
	}

	ginfo->io = data;
	ginfo->can_close = true;
	ioinfo_set_ready(ioinfo, ginfo->io);
	gensio_acc_free(accepter);
	if (debug)
		printf("Connected\r\n");

	oginfo->can_close = true;
	rv = gensio_open(oginfo->io, io_open, NULL);
	if (rv)
	{
		oginfo->can_close = false;
		fprintf(stderr, "Could not open %s: %s\n", oginfo->ios,
				gensio_err_to_str(rv));

		ginfo->err = rv;
		ginfo->o->wake(ginfo->waiter);
		return 0;
	}
	return 0;
}

#define MAX_CHANNELS 10
static const char *progname;

static void
help(int err)
{
	printf("%s [options]\n\n", progname);
	printf("A program to mux a number of tty's onto one gensio downstream device.\n");
	printf("This programs has many upstream tty, and one downstream gensios.\n\n");
	printf("options are:\n");
	printf("  -ds, --downstream <gensio> - One gensio device [serialdev, ...]\n");
	printf("  -us, --upstream <tty> - A semicolon separated list of tty devices\n");
	printf("  -d, --debug - Enable debug.  Specify more than once to increase\n"
		   "    the debug level\n");
	printf("  -a, --accepter - Accept a connection on downstream instead of"
		   " initiating a connection\n");
	printf("  -p, --printacc - When the accepter is started, print out all"
		   " the addresses being listened on.\n");
	printf("  -v, --verbose - Print all gensio logs\n");
	printf("  --signature <sig> - Set the RFC2217 server signature to <sig>\n");
	printf("  -e, --escchar - Set the local terminal escape character.\n"
		   "    Set to -1 to disable the escape character\n"
		   "    Default is ^\\ for tty stdin and disabled for non-tty stdin\n");
	printf("  -h, --help - This help\n");
	gensio_sel_exit(err);
}

static void
do_vlog(struct gensio_os_funcs *f, enum gensio_log_levels level,
		const char *log, va_list args)
{
	if (!debug)
		return;
	fprintf(stderr, "gensio %s log: ", gensio_log_level_to_str(level));
	vfprintf(stderr, log, args);
	fprintf(stderr, "\n");
}


static int
mux_event(struct gensio *io, void *user_data, int event, int ierr,
		  unsigned char *buf, gensiods *buflen,
		  const char *const *auxdata)
{
	switch (event)
	{
	case GENSIO_EVENT_READ:
	case GENSIO_EVENT_WRITE_READY:
		abort();

	case GENSIO_EVENT_NEW_CHANNEL:
		fprintf(stderr, "GENSIO_EVENT_NEW_CHANNEL");
		//new_rem_io((struct gensio *)buf, user_data);
		return 0;
	}

	return GE_NOTSUP;
}

static void
open_mux(struct gensio *io, struct gdata *ginfo)
{
	struct gensio_os_funcs *o = ginfo->o;
	struct gensio *mux_io;
	int err;
	static const char *isclient[2] = {"mode=server", NULL};

	err = mux_gensio_alloc(io, isclient, o, mux_event, ginfo, &mux_io);
	if (err)
	{
		fprintf(stderr, "Unable to allocate mux gensio: %s",
			   gensio_err_to_str(err));
		exit(1);
	}

	err = gensio_open_nochild_s(mux_io);
	if (err)
	{
		fprintf(stderr, "mux open failed: %s", gensio_err_to_str(err));
		exit(1);
	}

	start_local_ports(mux_io);
//	new_rem_io(mux_io, ginfo);
}

int main(int argc, char *argv[])
{
	int arg, rv;
	struct gensio_waiter *closewaiter = NULL;
	unsigned int closecount = 0;
//	bool downstream_do_acc = false, downstream_do_print = false;
//	struct gensio_accepter *ds_acc = NULL;
	bool esc_set = false;
	bool us_set = false;
	bool ds_set = false;
	int escape_char = -1;
	char *signature = "gensiomux";
	char *us_ttys = NULL;
	//char *us_tty[MAX_CHANNELS] = {NULL};
	char *ds_gensio = NULL;
	struct gensio_os_funcs *o = NULL;
	struct ioinfo_sub_handlers *us_sh = NULL, *ds_sh = NULL;
	void *us_subdata = NULL, *ds_subdata = NULL;
	struct ioinfo *us_ioinfo = NULL, *ds_ioinfo = NULL;
//	struct gdata us_data;
	struct gdata ds_data;
	char *filename;
	sigset_t sigs;
	struct timeval zerotime = {0, 0};

	/*
     * Make sure that SIGPIPE doesn't kill is if the user is doing
     * something involving a pipe.
     */
	sigemptyset(&sigs);
	sigaddset(&sigs, SIGPIPE);
	sigaddset(&sigs, SIGHUP);
	rv = sigprocmask(SIG_BLOCK, &sigs, NULL);
	if (rv)
	{
		perror("Could not set up signal mask");
		exit(1);
	}

//	memset(&us_data, 0, sizeof(us_data));
	memset(&ds_data, 0, sizeof(ds_data));
	gensio_list_init(&ds_data.cons);

	progname = argv[0];

	for (arg = 1; arg < argc; arg++)
	{
		if (argv[arg][0] != '-')
			break;
		if (strcmp(argv[arg], "--") == 0)
		{
			arg++;
			break;
		}
		fprintf(stderr, "Argument: %s\n", argv[arg]);

		if ((rv = cmparg(argc, argv, &arg, "-us", "--upstream", &us_ttys)) != 0)
			us_set = true;
		else if ((rv = cmparg(argc, argv, &arg, "-ds", "--downstream", &ds_gensio)) != 0)
			ds_set = true;
//		else if ((rv = cmparg(argc, argv, &arg, "-a", "--accepter", NULL)) != 0)
//			downstream_do_acc = true;
//		else if ((rv = cmparg(argc, argv, &arg, "-p", "--printacc", NULL)) != 0)
//			downstream_do_print = true;
		else if ((rv = cmparg(argc, argv, &arg, "-v", "--verbose", NULL)) != 0)
			gensio_set_log_mask(GENSIO_LOG_MASK_ALL);
		else if ((rv = cmparg_int(argc, argv, &arg, "-e", "--escchar", &escape_char)) != 0)
			esc_set = true;
		else if ((rv = cmparg(argc, argv, &arg, "", "--signature", &signature)) != 0)
			; // nop
		else if ((rv = cmparg(argc, argv, &arg, "-d", "--debug", NULL)) != 0)
		{
			debug++;
			if (debug > 1)
				gensio_set_log_mask(GENSIO_LOG_MASK_ALL);
		}
		else if ((rv = cmparg(argc, argv, &arg, "-h", "--help", NULL)) != 0)
			help(0);
		else if ((rv = cmparg(argc, argv, &arg, NULL, "--dummyrand", &filename)) != 0)
		{
#ifdef HAVE_OPENSSL
			/*
	     * This option is undocumented and only for testing.  Do not
	     * use it!
	     */
			if (dummyrnd_file != -1)
				close(dummyrnd_file);
			dummyrnd_file = open(filename, O_RDONLY);
			if (dummyrnd_file == -1)
			{
				fprintf(stderr, "Could not open rand file: %s\n",
						strerror(errno));
				goto out_err;
			}

			rv = RAND_set_rand_method(&dummyrnd);
			if (rv != 1)
			{
				fprintf(stderr, "Error setting random method\n");
				goto out_err;
			}
#endif
		}
		else
		{
			fprintf(stderr, "Unknown argument: %s\n", argv[arg]);
			help(1);
		}
		if (rv < 0)
			goto out_err;
	}

	if (us_set && !esc_set)
		escape_char = 0; /* disable */

//	if (arg >= argc)
    if ((us_ttys == NULL) || (ds_gensio == NULL))
	{
		fprintf(stderr, "No gensio string given to connect to\n");
		help(1);
	}

	// Assign OS handler to upstream and downstream gensio
	if (rv = gensio_default_os_hnd(0, &o) != 0)
	{
		fprintf(stderr, "Could not allocate OS handler: %s\n",
				gensio_err_to_str(rv));
		goto out_err;
	}
	o->vlog = do_vlog;
//	us_data.o = o;
	ds_data.o = o;

	// Break upstream tty list into tokens, and create local ports
	char* tty_rest = us_ttys;
	char* tty_token;
	while ((tty_token = strtok_r(tty_rest, ",", &tty_rest)) != NULL)
	{
		char tty[50];
		char tty_gensio[50];
		char tty_service[50];
		char* p;
		strcpy(tty, tty_token);

		// Create config string for upstream TTY gensio  config string
		p = strtok(tty, "/");
		snprintf(tty_gensio, sizeof(tty_gensio)-1, "serialdev,/dev/%s,115200N81", p);

		// Create config string for upstream TTY service name
		p = strtok(tty, "/");
		snprintf(tty_service, sizeof(tty_service)-1, "%s", p);

		if ((rv = add_local_port(o, tty_gensio, tty_service, tty_gensio)) != 0)
		{
			fprintf(stderr, "Could not add local port: %s\n",
				gensio_err_to_str(rv));
			goto out_err;
		}
	}

//	us_data.ios = us_tty[0];	// XXXMAW For single instance

	ds_data.ios = ds_gensio;

	// Assign OS waiter to upstream and downstream gensio
	if ((ds_data.waiter = o->alloc_waiter(o)) == 0)
	{
		ds_data.err = GE_NOMEM;
		fprintf(stderr, "Could not allocate OS waiter\n");
		goto out_err;
	}
	//us_data.waiter = ds_data.waiter;

	if ((closewaiter = o->alloc_waiter(o)) == 0)
	{
		//us_data.err = GE_NOMEM;
		ds_data.err = GE_NOMEM;
		fprintf(stderr, "Could not allocate close waiter\n");
		goto out_err;
	}

    // Allocate upstream and downstream serial gensio sub-handler
	/*
	if ((us_subdata = alloc_ser_ioinfo(0, signature, &us_sh)) == 0)
	{
		us_data.err = GE_NOMEM;
		fprintf(stderr, "Could not allocate upstream subdata\n");
		goto out_err;
	}
	*/
	if ((ds_subdata = alloc_ser_ioinfo(0, signature, &ds_sh)) == 0)
	{
		ds_data.err = GE_NOMEM;
		fprintf(stderr, "Could not allocate downstream subdata\n");
		goto out_err;
	}

    // Allocate upstream and downstream ioinfo
	/*
	if ((us_ioinfo = alloc_ioinfo(o, escape_char, us_sh, us_subdata, &guh, &us_data)) == 0)
	{
		us_data.err = GE_NOMEM;
		fprintf(stderr, "Could not allocate upstream ioinfo\n");
		goto out_err;
	}
	*/
	if ((ds_ioinfo = alloc_ioinfo(o, -1, ds_sh, ds_subdata, &guh, &ds_data)) == 0)
	{
		//us_data.err = GE_NOMEM;
		ds_data.err = GE_NOMEM;
		fprintf(stderr, "Could not allocate downstream ioinfo\n");
		goto out_err;
	}

	// Connect upstream and downstream ioinfo
	/*
	ioinfo_set_otherioinfo(us_ioinfo, ds_ioinfo);
	*/

/*
	if ((rv = str_to_gensio(us_data.ios, o, NULL, us_ioinfo, &us_data.io)) != 0)
	{
		fprintf(stderr, "Could not allocate %s: %s\n",
				us_data.ios, gensio_err_to_str(rv));
		us_data.err = rv;
		goto out_err;
	}

	us_data.user_io = us_data.io;
	ds_data.user_io = us_data.io;  // <<<<<<<<<<<<<<<<<<<
*/
#if 0
	if (downstream_do_acc)
	{
		/*
		 * Create downstream accepter gensio
		 */
		rv = str_to_gensio_accepter(ds_data.ios, ds_data.o, io_acc_event, ds_ioinfo, &ds_acc);
		if (rv)
		{
			ds_data.err = rv;
			fprintf(stderr, "Could not allocate accepter %s: %s\n", ds_data.ios,
				gensio_err_to_str(rv));
			goto out_err;
		}
	}
	else
#endif
	{
		/*
		 * Create downstream connector gensio
		 */
		rv = str_to_gensio(ds_data.ios, ds_data.o, NULL, ds_ioinfo, &ds_data.io);
		if (rv)
		{
			ds_data.err = rv;
			fprintf(stderr, "Could not allocate %s: %s\n", ds_data.ios,
				gensio_err_to_str(rv));
			goto out_err;
		}
	gensio_set_is_reliable(ds_data.io, true);
	}

#if 0
	if (downstream_do_acc)
	{
		/*
		 * Start downstream gensio as accepter
		 */
		if ((rv = gensio_acc_startup(ds_acc)) != 0)
		{
			fprintf(stderr, "Could not start %s: %s\n", ds_data.ios,
					gensio_err_to_str(rv));
			ds_data.err = rv;
			goto close1;
		}
		/*
		if (downstream_do_print)
		{
			char str[2048];
			gensiods size;
			unsigned int i;

			for (i = 0;; i++)
			{
				snprintf(str, sizeof(str), "%u", i);
				size = sizeof(str);
				rv = gensio_acc_control(ds_acc, GENSIO_CONTROL_DEPTH_FIRST,
										true, GENSIO_ACC_CONTROL_LADDR,
										str, &size);
				if (rv == GE_NOTFOUND)
					break;
				if (rv)
					fprintf(stderr,
							"Unable to fetch accept address %d: %s\n", i,
							gensio_err_to_str(rv));
				else
					fprintf(stderr, "Address %d: %s\n", i, str);
			}
			fprintf(stderr, "Done\n");
		}
		*/
	}
	else
#endif
	{
		/*
		 * Start downstream gensio as connector
		 */
		ds_data.can_close = true;

		if ((rv = gensio_open(ds_data.io, io_open, NULL)) > 0)
		{
			ds_data.err = rv;
			ds_data.can_close = false;
			fprintf(stderr, "Could not open %s: %s\n", ds_data.ios,
					gensio_err_to_str(rv));
			goto close1;
		}

		/* At this point we are fully authenticated and have all global info. */

		open_mux(ds_data.io, &ds_data);

		/*
		us_data.can_close = true;

		if ((rv = gensio_open(us_data.io, io_open, NULL)) > 0)
		{
			us_data.err = rv;
			us_data.can_close = false;
			fprintf(stderr, "Could not open %s: %s\n", us_data.ios,
					gensio_err_to_str(rv));
			goto out_err;
		}
		*/
	}

	/*
	 * Go process stream data
	 */
	ds_data.o->wait(ds_data.waiter, 1, NULL);

out_err:
	if (ds_data.can_close)
	{
		rv = gensio_close(ds_data.io, io_close, closewaiter);
		if (rv)
			fprintf(stderr, "Unable to close %s: %s\n", ds_data.ios,
					gensio_err_to_str(rv));
		else
			closecount++;
	}
	/*
	else if (!ds_data.io && downstream_do_acc && ds_acc)
	{
		gensio_acc_free(ds_acc);
	}
*/

close1:
/*
	if (us_data.can_close)
	{
		rv = gensio_close(us_data.io, io_close, closewaiter);
		if (rv)
			fprintf(stderr, "Unable to close %s: %s\n", us_data.ios,
					gensio_err_to_str(rv));
		else
			closecount++;
	}
*/
	if (closecount > 0)
	{
		o->wait(closewaiter, closecount, NULL);
	}

/*
	if (us_data.io)
		gensio_free(us_data.io);
		*/
	if (ds_data.io)
		gensio_free(ds_data.io);

	if (us_ioinfo)
		free_ioinfo(us_ioinfo);
	if (ds_ioinfo)
		free_ioinfo(ds_ioinfo);
	if (us_subdata)
		free_ser_ioinfo(us_subdata);
	if (ds_subdata)
		free_ser_ioinfo(ds_subdata);

/*
	if (!rv && us_data.err)
		rv = us_data.err;
		*/
	if (!rv && ds_data.err)
		rv = ds_data.err;

	while (o && o->service(o, &zerotime) == 0)
		;
/*
	if (us_data.waiter)
		o->free_waiter(us_data.waiter);
		*/
	if (closewaiter)
		o->free_waiter(closewaiter);
	if (o)
		gensio_cleanup_mem(o);
	gensio_sel_exit(!!rv);
}
